#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>

// 

//  
HWND               hWnd       = NULL;

//   Direct3D
IDirect3D9*        pD3D       = NULL;

//  
IDirect3DDevice9*  pDevice = NULL;

//   
IDirect3DVertexBuffer9* vb = NULL;

//   
IDirect3DIndexBuffer9*  ib = NULL;

//  
float SpinX = 0.0f;
float SpinY = 0.0f;

// 
bool Flat = false;

//,   
struct Vertex
{
	float x, y, z;
	DWORD c;

	static const DWORD FVF;
};

const DWORD Vertex::FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE;

//   
Vertex Vertices[] =
{
	{-1.0f, 1.0f,-1.0f,  D3DCOLOR_COLORVALUE( 1.0, 0.0, 0.0, 1.0 ) }, // 0
	{ 1.0f, 1.0f,-1.0f,  D3DCOLOR_COLORVALUE( 0.0, 1.0, 0.0, 1.0 ) }, // 1
	{-1.0f,-1.0f,-1.0f,  D3DCOLOR_COLORVALUE( 0.0, 0.0, 1.0, 1.0 ) }, // 2
	{ 1.0f,-1.0f,-1.0f,  D3DCOLOR_COLORVALUE( 1.0, 1.0, 0.0, 1.0 ) }, // 3
	{-1.0f, 1.0f, 1.0f,  D3DCOLOR_COLORVALUE( 1.0, 0.0, 1.0, 1.0 ) }, // 4
	{-1.0f,-1.0f, 1.0f,  D3DCOLOR_COLORVALUE( 0.0, 1.0, 1.0, 1.0 ) }, // 5
	{ 1.0f, 1.0f, 1.0f,  D3DCOLOR_COLORVALUE( 1.0, 1.0, 1.0, 1.0 ) }, // 6
	{ 1.0f,-1.0f, 1.0f,  D3DCOLOR_COLORVALUE( 1.0, 0.0, 0.0, 1.0 ) }  // 7
};

//     
WORD Indices[] =
{
	0, 1, 2, 3, // 0
	4, 5, 6, 7, // 1
	4, 6, 0, 1, // 2
	5, 2, 7, 3, // 3
	1, 6, 3, 7, // 4
	0, 2, 4, 5  // 5
};

//  
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void Init(void);
void Shutdown(void);
void Render(void);


//

//  
int WINAPI WinMain( HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR     lpCmdLine,
					int       nCmdShow)
{
	// 
	WNDCLASS wc;

	// 
	MSG        Msg;
    
	//   
	wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(0, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(0, IDC_ARROW);
    wc.hbrBackground = static_cast<HBRUSH>(GetStockObject(BLACK_BRUSH));
    wc.lpszMenuName  = 0;
    wc.lpszClassName = "Cube";

	//    - 
	if(!RegisterClass(&wc))
	{
		MessageBox(0, "   .", 0, 0);
        exit(1);
	}
	
	//  
	hWnd = CreateWindowEx(WS_EX_OVERLAPPEDWINDOW,
						  "Cube", 
                          "Direct3D (DX9) - Cube",
						  WS_OVERLAPPEDWINDOW,
					      0, 0, 640, 480, NULL, NULL, hInstance, NULL );

	//    - 
	if(hWnd == NULL)
	{
		MessageBox(0, "  ", 0, 0);
        exit(1);
	}
		
	//   
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

	//  
	ZeroMemory(&Msg, sizeof(Msg));
 
	// Direct3D
	Init();

	//      
	while(Msg.message != WM_QUIT)
	{
		//   ,
		if(PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE))
		{ 
			//    
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}

		//
		Render();
	}

	// 
	Shutdown();

	//  
    UnregisterClass("Cube", wc.hInstance);

	return Msg.wParam;
}

//  ,   
// 
LRESULT CALLBACK WndProc(HWND   hWnd, 
						 UINT   msg, 
						 WPARAM wParam, 
						 LPARAM lParam)
{
	static POINT LastMPos;
	static POINT CurrMPos;
	static bool lButton;

	switch(msg)
	{
	case WM_KEYDOWN:
		{
			switch(wParam)
			{
			case VK_ESCAPE:
				PostQuitMessage(0);
				break;
			//F1, F2 -   
			case VK_F1:
				Flat = true;
				break;

			case VK_F2:
				Flat = false;
				break;
			}
		}
		break;

	case WM_LBUTTONDOWN:
		{
			LastMPos.x = CurrMPos.x = LOWORD (lParam);
			LastMPos.y = CurrMPos.y = HIWORD (lParam);
			lButton = true;
		}
		break;

	case WM_LBUTTONUP:
		{
			lButton = false;
		}
		break;

	case WM_MOUSEMOVE:
		{
			CurrMPos.x = LOWORD (lParam);
			CurrMPos.y = HIWORD (lParam);

			if(lButton)
			{
				SpinX -= (CurrMPos.x - LastMPos.x);
				SpinY -= (CurrMPos.y - LastMPos.y);
			}

			LastMPos.x = CurrMPos.x;
			LastMPos.y = CurrMPos.y;
		}
		break;

		case WM_CLOSE:
		{
			PostQuitMessage(0);	
		}
		
        case WM_DESTROY:
		{
            PostQuitMessage(0);
		}
        break;

		default:
		{
			return DefWindowProc( hWnd, msg, wParam, lParam );
		}
		break;
	}

	return 0;
}

//  Direct3D,  
//  ,    
void Init(void)
{
	//CAPS 
	D3DCAPS9 caps;
	
	//   
	pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	
	pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT,	D3DDEVTYPE_HAL,	&caps);	
	
	//   
	int vp = 0;
	if(caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
		vp = D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		vp = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	//   .
	D3DPRESENT_PARAMETERS d3dpp = {0};
	d3dpp.BackBufferWidth            = 640;
	d3dpp.BackBufferHeight           = 480;
	d3dpp.BackBufferFormat           = D3DFMT_A8R8G8B8; // 
	d3dpp.BackBufferCount            = 1;
	d3dpp.MultiSampleType            = D3DMULTISAMPLE_NONE;
	d3dpp.MultiSampleQuality         = 0;
	d3dpp.SwapEffect                 = D3DSWAPEFFECT_DISCARD;
	d3dpp.hDeviceWindow              = hWnd;
	d3dpp.Windowed                   = true; 
	d3dpp.EnableAutoDepthStencil     = true;
	d3dpp.AutoDepthStencilFormat     = D3DFMT_D24S8; //   
	d3dpp.Flags                      = 0;
	d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	d3dpp.PresentationInterval       = D3DPRESENT_INTERVAL_IMMEDIATE;

	//  
	pD3D->CreateDevice(D3DADAPTER_DEFAULT, 
					   D3DDEVTYPE_HAL, 
					   hWnd,
                       D3DCREATE_HARDWARE_VERTEXPROCESSING,
                       &d3dpp, 
					   &pDevice);
	
	// 
	pDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	// Z - 
	pDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

	//  
	D3DXMATRIX matProj;
	D3DXMatrixPerspectiveFovLH(&matProj, D3DXToRadian(45.0f), 
                                640.0f / 480.0f, 0.1f, 100.0f);
	pDevice->SetTransform( D3DTS_PROJECTION, &matProj );


	//  
	pDevice->CreateVertexBuffer(8*sizeof(Vertex),
		                        D3DUSAGE_WRITEONLY,
								Vertex::FVF,
		                        D3DPOOL_DEFAULT,
								&vb,
								NULL);
	WORD *pVertices = NULL;

	// 
	vb->Lock(0, sizeof(Vertices), (void**)&pVertices, 0);
	//     
	memcpy(pVertices, Vertices, sizeof(Vertices));
	// 
	vb->Unlock();

	//  
	pDevice->CreateIndexBuffer(24*sizeof(WORD),
							   D3DUSAGE_WRITEONLY,
							   D3DFMT_INDEX16,
							   D3DPOOL_DEFAULT,
							   &ib,
							   NULL);
	WORD *pIndices = NULL;

	// 
	ib->Lock(0, sizeof(Indices), (void**)&pIndices, 0);
	//     
	memcpy(pIndices, Indices, sizeof(Indices));
	// 
	ib->Unlock();
}

//  . 
//     ,
//  ""
void Shutdown(void)
{

	if(vb != NULL)
		vb->Release(); 

	if(ib != NULL)
		ib->Release(); 

    if(pDevice != NULL)
        pDevice->Release();

    if(pD3D != NULL)
        pD3D->Release();
}

//  
void Render(void)
{
	//   Z -      
    pDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                         D3DCOLOR_COLORVALUE(0.0f,0.0f,0.0f,1.0f), 1.0f, 0);

	
	// 
	D3DXMATRIX matTrans;
	// 
	D3DXMATRIX matRot;
	// 
	D3DXMATRIX matWorld;

	//  
	D3DXMatrixTranslation(&matTrans, 0.0f, 0.0f, 5.0f);

	//  
	D3DXMatrixRotationYawPitchRoll(&matRot,
		                           D3DXToRadian(SpinX),
		                           D3DXToRadian(SpinY),
		                           0.0f);
	//  
	matWorld = matRot * matTrans;
	pDevice->SetTransform(D3DTS_WORLD, &matWorld);

	if(Flat)
		//  
		pDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT); 
	else
		//   
		pDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);

	// 
    pDevice->BeginScene();

	//     
	pDevice->SetStreamSource(0, vb, 0, sizeof(Vertex));
	pDevice->SetIndices(ib);
	//   
	pDevice->SetFVF(Vertex::FVF);

	// 
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8,  0, 2);
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8,  4, 2);
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8,  8, 2);
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8, 12, 2);
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8, 16, 2);
    pDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0, 0, 8, 20, 2);

	// 
    pDevice->EndScene();
	//  
    pDevice->Present(NULL, NULL, NULL, NULL);
}
